//
// Copyright (c) 2015-2019 CNRS INRIA
// Copyright (c) 2015 Wandercraft, 86 rue de Paris 91400 Orsay, France.
//

#ifndef __pinocchio_multibody_joint_base_hpp__
#define __pinocchio_multibody_joint_base_hpp__

#include "pinocchio/multibody/joint/fwd.hpp"

#include "pinocchio/multibody/joint/joint-model-base.hpp"
#include "pinocchio/multibody/joint/joint-data-base.hpp"

#endif // ifndef __pinocchio_multibody_joint_base_hpp__
