//
// Copyright (c) 2018 CNRS
//

#ifndef __pinocchio_lie_group_fwd_hpp__
#define __pinocchio_lie_group_fwd_hpp__

#include "pinocchio/fwd.hpp"

namespace pinocchio
{
  template<typename LieGroupCollection> struct LieGroupGenericTpl;
}

#endif // ifndef __pinocchio_lie_group_fwd_hpp__
